from lstmcpipe.config.paths_config import PathConfigAllSkyFullDL1ab

prod_id = '20220518_allsky_dec2276_tuned'

cfg = PathConfigAllSkyFullDL1ab('20220511_allsky_std', prod_id, ['dec_2276'])

cfg.generate()

cfg.save_yml(f'lstmcpipe_config_{prod_id}.yml')



"""
Run python rerun_irfs_mem_issue.py lstmcpipe_config.yaml
"""
from ruamel.yaml import YAML
import sys
from pathlib import Path


yaml = YAML()

filename = Path(sys.argv[1])

cfg = yaml.load(open(filename).read())

cfg['stages_to_run'] = ['dl2_to_irfs']
cfg['stages'].pop('r0_to_dl1', None)
cfg['stages'].pop('merge_dl1', None)
cfg['stages'].pop('train_pipe', None)
cfg['stages'].pop('dl1_to_dl2', None)

to_rerun = []
for p in cfg['stages']['dl2_to_irfs']:
    if not Path(p['output']).is_file():
        print(p['output'])
        p['slurm_options'] = '--mem=6GB'
        to_rerun.append(p)

cfg['stages']['dl2_to_irfs'] = to_rerun
print(f"{len(to_rerun)} irfs to re-produce")


output_filename = filename.parent.joinpath(filename.stem + '_rerun.yaml')

if to_rerun and not Path(output_filename).exists():
    print(f"Run lstmcpipe -c {output_filename}")
    yaml.dump(cfg, open(output_filename, 'w'))
elif Path(output_filename).exists():
    raise FileExistsError(output_filename)



from pathlib import Path
from ruamel.yaml import YAML
import json


def codemeta2citation(codemeta_path='codemeta.json', citation_path='CITATION.cff'):
    codemeta = json.load(open(codemeta_path))
    citation = {'title': codemeta['name'],
                'type': 'software',
                'authors': [],
                'message': 'If you use this software, please cite it using Zenodo from https://doi.org/10.5281/zenodo.6460727',
                'license': 'MIT',
                }
    for author in codemeta['author']:
        citation['authors'].append({
            'family-names': author['familyName'],
            'given-names': author['givenName'],
            'orcid': author['@id'],
            'email': author['email']
        }
        )

    yaml = YAML()
    yaml.dump(citation, open(citation_path, 'w'))


if __name__ == '__main__':
    codemeta_path = Path(__file__).parents[3].joinpath('codemeta.json')
    citation_path = Path(__file__).parents[3].joinpath('CITATION.cff')
    codemeta2citation(codemeta_path, citation_path)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.


import os
import sys
import json
import datetime

sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

with open(os.path.join(os.path.dirname(__file__), "..", "codemeta.json")) as file:
    metadata = json.load(file)

# General information about the project.
project = metadata["name"]
author = ""
for aut in metadata["author"]:
    author += f"{aut['givenName']} {aut['familyName']},"

copyright = "{}.  Last updated {}".format(author, datetime.datetime.now().strftime("%d %b %Y %H:%M"))

# The full version, including alpha/beta/rc tags
release = metadata['version']

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_automodapi.automodapi",
    "sphinxarg.ext",
    "sphinx.ext.napoleon",
    "sphinxcontrib.mermaid",
]

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



#!/usr/bin/env python

import os
import json
import shutil
import logging
import warnings
import subprocess as sp
from pathlib import Path
from ruamel.yaml import YAML
from pprint import pprint
from copy import deepcopy
from deepdiff import DeepDiff


log = logging.getLogger(__name__)


def save_log_to_file(dictionary, output_file, workflow_step=None):
    """
    Dumps a dictionary (log) into a dicts of dicts with keys each of the pipeline stages.

    Parameters
    ----------
    dictionary : dict
        The dictionary to be dumped to a file
    output_file : str or Path
        Output file to store the log
    workflow_step : str
        Step of the workflow, to be recorded in the log

    Returns
    -------
        None
    """
    if workflow_step is None:
        workflow_step = "NoKEY"

    dict2log = {workflow_step: dictionary}

    with open(output_file, "a+") as fileout:
        YAML().dump(dict2log, fileout)


def batch_mc_production_check(
    dict_jobids_all_stages,
    log_directory,
    prod_id,
    prod_config_file,
    batch_config,
    logs_files,
):
    """
    Check that the dl1_to_dl2 stage, and therefore, the whole workflow has ended correctly.
    The machine information of each job will be dumped to the file.
    The file will take the form `check_MC_prodID_{prod_id}_OK.txt`

    Parameters
    ----------
    dict_jobids_all_stages : dict
        dict containing the {stage: all_job_ids related} information
    log_directory: Path
    prod_id: str
    prod_config_file: str
    batch_config: dict
    logs_files: dict
        Dictionary with logs files

    Returns
    -------
    jobid : str

    """
    debug_log = {}
    all_pipeline_jobs = []

    source_env = batch_config["source_environment"]
    slurm_account = batch_config["slurm_account"]

    for stage, jobids in dict_jobids_all_stages.items():
        all_pipeline_jobs.append(jobids)
        debug_log.update({f"SUMMARY_{stage}": jobids})

    all_pipeline_jobs = ",".join(all_pipeline_jobs)

    # Copy lstmcpipe config used to log directory
    shutil.copyfile(Path(prod_config_file).resolve(), log_directory.joinpath(f"config_MC_prod_{prod_id}.yml"))

    # Save machine info into the check file
    check_prod_file = log_directory.joinpath(f"check_MC_{prod_id}.txt").absolute().as_posix()

    cmd_wrap = f"touch {check_prod_file}; "
    cmd_wrap += (
        f"sacct --format=jobid,jobname,nodelist,cputime,state,exitcode,avediskread,maxdiskread,avediskwrite,"
        f"maxdiskwrite,AveVMSize,MaxVMSize,avecpufreq,reqmem -j {all_pipeline_jobs} >> {check_prod_file}; "
        f"mv slurm-* IRFFITSWriter.provenance.log {log_directory.absolute().as_posix()} "
    )

    batch_cmd = "sbatch -p short --parsable"
    if slurm_account != "":
        batch_cmd += f" -A {slurm_account}"
    batch_cmd += f" --dependency=afterok:{all_pipeline_jobs} -J prod_check" f' --wrap="{source_env} {cmd_wrap}"'

    jobid = os.popen(batch_cmd).read().strip("\n")
    log.info(f"Submitted batch CHECK-job {jobid}")
    debug_log.update({f"prod_check_{jobid}": batch_cmd})

    save_log_to_file(debug_log, logs_files["debug_file"], workflow_step="check_full_workflow")

    return jobid


def rerun_cmd(cmd, outfile, max_ntry=2, subdir_failures='failed_outputs', **run_kwargs):
    """
    rerun r0_to_dl1 process given by `cmd` as long as the exit code is 0 and number of try < max_ntry
    move the failed output file to subdir failed_outputs

    Parameters
    ----------
    cmd: str
    subdir_failures: str
    outfile: Path
        path to the cmd output file
    max_ntry: int
    run_kwargs: kwargs for subprocess.run

    Returns
    -------
    ntry: int
        number of tries actually run
    """
    outfile = Path(outfile)
    ret = -1
    ntry = 1
    while ret != 0 and ntry <= max_ntry:
        ret = sp.run(cmd, **run_kwargs).returncode
        if ret != 0:
            failed_jobs_subdir = outfile.parent.joinpath(subdir_failures)
            if outfile.exists():
                failed_jobs_subdir.mkdir(exist_ok=True)
                outfile_target = failed_jobs_subdir.joinpath(outfile.name)
                print(f"Move failed output file from {outfile} to {outfile_target}. try #{ntry}")
                shutil.move(outfile, outfile_target)

        ntry += 1
    return ntry - 1


def dump_lstchain_std_config(filename='lstchain_config.json', allsky=False, overwrite=False):
    from lstchain.io.config import get_standard_config

    filename = Path(filename)

    if filename.exists() and not overwrite:
        raise FileExistsError(f"{filename} exists already")

    std_cfg = get_standard_config()
    cfg = deepcopy(std_cfg)

    cfg['LocalPeakWindowSum']['apply_integration_correction'] = True
    cfg['GlobalPeakWindowSum']['apply_integration_correction'] = True
    cfg['source_config']['EventSource']['allowed_tels'] = [1]
    cfg['random_forest_energy_regressor_args']['min_samples_leaf'] = 10
    cfg['random_forest_disp_regressor_args']['min_samples_leaf'] = 10
    cfg['random_forest_disp_classifier_args']['min_samples_leaf'] = 10
    cfg['random_forest_particle_classifier_args']['min_samples_leaf'] = 10
    cfg['random_forest_energy_regressor_args']['n_jobs'] = -1
    cfg['random_forest_disp_regressor_args']['n_jobs'] = -1
    cfg['random_forest_disp_classifier_args']['n_jobs'] = -1
    cfg['random_forest_particle_classifier_args']['n_jobs'] = -1

    if allsky:
        for rf_feature in [
            'energy_regression_features',
            'disp_regression_features',
            'disp_classification_features',
            'particle_classification_features',
        ]:
            cfg[rf_feature] = std_cfg[rf_feature]
            if 'alt_tel' not in cfg[rf_feature]:
                cfg[rf_feature].append('alt_tel')
            if 'az_tel' not in cfg[rf_feature]:
                cfg[rf_feature].append('az_tel')

    extra_msg = "for AllSky prod" if allsky else ""
    print(f"Updating std lstchain config {extra_msg} with")
    diff = DeepDiff(std_cfg, cfg)
    pprint(diff)

    with open(filename, 'w') as file:
        json.dump(cfg, file, indent=4)
    print(f"\nModified lstchain config dumped in {filename}. Check full config thoroughly.")


def run_command(*args):
    """
    Runs the command passed through args, as a subprocess.Popen() call.

    Based on:
    https://github.com/cta-observatory/cta-lstchain/blob/master/lstchain/scripts/tests/test_lstchain_scripts.py#L43

    Parameters
    ----------
    args: str or iter
        `Shell` is forced to True, thus a single string (shell cmd) is expected.

    Returns
    -------
    (subprocess.Popen.stdout).strip('\n')
    """
    cmd = sp.Popen(args, shell=True, stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf-8')
    stdout, _ = cmd.communicate()

    if cmd.returncode != 0:
        raise ValueError(f"Running `{args[0]}` failed with return code {cmd.returncode}, output: \n {stdout}")
    else:
        return stdout.strip('\n')


class SbatchLstMCStage:
    """
    Base class to (slurm) sbatch a lstMCpipe stage
    """

    def __init__(
        self,
        stage,
        wrap_command,
        slurm_output=None,
        slurm_error=None,
        job_name=None,
        slurm_account=None,
        slurm_partition=None,
        slurm_deps=None,
        slurm_options=None,
        source_environment="",
        backend="",
    ):
        self.base_slurm_command = "sbatch --parsable"

        self.slurm_output = f"--output={slurm_output}" if slurm_output is not None else "--output=./slurm-%j.o"
        self.slurm_error = f"--error={slurm_error}" if slurm_error is not None else "--error=./slurm-%j.e"
        self.job_name = f"--job-name={job_name}" if job_name is not None else ""
        self.slurm_account = f"--account={slurm_account}" if slurm_account is not None else ""

        self.slurm_partition = f"--partition={slurm_account}" if slurm_partition is not None else "--partition=short"
        self.slurm_options = f"{slurm_options}" if slurm_options is not None else None

        self.stage_default_options(stage)
        self.slurm_dependencies = None
        self.check_slurm_dependencies(slurm_deps)
        self.wrap_cmd = wrap_command
        self.compose_wrap_command(wrap_command, source_environment, backend)

    def __str__(self):
        # return full string
        return self.slurm_command

    @property
    def _valid_stages(self):
        return [
            "r0_to_dl1",
            "dl1ab",
            "merge_dl1",
            "train_test_splitting",
            "train_pipe",
            "RF_importance",
            "dl1_to_dl2",
            "dl2_to_irfs",
            "dl2_sens",
            "dl2_sens_plot",
        ]

    def compose_wrap_command(self, wrap_command=None, source_env="", backend=""):
        if wrap_command is None or wrap_command == "":
            warnings.warn("You must pass a command to be batched! ")
        if source_env != "" and not source_env.strip().endswith(";"):
            source_env = source_env.strip() + "; "
        if backend != "" and not backend.strip().endswith(";"):
            backend = backend.strip() + "; "
        self.wrap_cmd = f'--wrap="{backend}{source_env}{wrap_command}"'

    @property
    def slurm_command(self):
        if self.slurm_options is not None:
            return (
                f"{self.base_slurm_command} {self.job_name} {self.slurm_options}"
                f" {self.slurm_error} {self.slurm_output} {self.slurm_dependencies}"
                f" {self.slurm_account} {self.wrap_cmd}"
            )
        else:
            return (
                f"{self.base_slurm_command} {self.job_name} {self.slurm_partition}"
                f" {self.slurm_error} {self.slurm_output} {self.slurm_dependencies}"
                f" {self.slurm_account} {self.wrap_cmd}"
            )

    def check_slurm_dependencies(self, slurm_deps, dependency="afterok"):
        if slurm_deps is None:
            self.slurm_dependencies = ""
        else:
            if all(items != "" for items in slurm_deps.split(",")):
                self.slurm_dependencies = f"--dependency={dependency}:{slurm_deps}"
            else:
                raise ValueError("Slurm dependencies contain an empty value between commas, i.e.; ,'', ")

    def stage_default_options(self, stage):
        if stage not in self._valid_stages or stage is None:
            raise ValueError(f"Please select a valid stage: \n{', '.join(self._valid_stages)}")

        _default_options = {
            "r0_to_dl1": getattr(self, "r0_dl1_options"),
            "dl1ab": getattr(self, "dl1ab_options"),
            "merge_dl1": getattr(self, "set_merge_dl1_default_options"),
            "train_test_splitting": getattr(self, "set_train_test_splitting_default_options"),
            "train_pipe": getattr(self, "set_trainpipe_default_options"),
            "RF_importance": getattr(self, "set_train_plot_rf_feat_default_options"),
            "dl1_to_dl2": getattr(self, "set_dl1_dl2_default_options"),
            "dl2_to_irfs": getattr(self, "set_dl2_irfs_default_options"),
            "dl2_sens": getattr(self, "set_dl2_sens_default_options"),
            "dl2_sens_plot": getattr(self, "set_dl2_sens_plot_default_options"),
        }
        _default_options[stage]()

    def submit(self):
        if self.wrap_cmd is None or self.wrap_cmd == "":
            raise ValueError(
                "You must first define the command to be batched: " "SbatchLstMCStage().wrap_command('COMMAND')"
            )
        else:
            jobid = run_command(self.slurm_command)
            return jobid

    def r0_dl1_options(self, process_dl1_job_name="r0_dl1", array="0-0%100", partition="long", extra_slurm_options=""):
        self.job_name = f"--job-name={process_dl1_job_name}"
        self.slurm_options = f"--partition={partition} --array={array} {extra_slurm_options}"

    def dl1ab_options(self, process_dl1ab_job_name="dl1ab", array="0-0%100", partition="long", extra_slurm_options=""):
        self.job_name = f"--job-name={process_dl1ab_job_name}"
        self.slurm_options = f"--partition={partition} --array={array} {extra_slurm_options}"

    def set_merge_dl1_default_options(self):
        self.job_name = "--job-name=merge"
        self.slurm_partition = "--partition=long"

    def set_train_test_splitting_default_options(self):
        self.job_name = "--job-name=train_test_splitting"
        self.slurm_partition = "--partition=short"

    def set_trainpipe_default_options(self):
        self.job_name = "--job-name=train_pipe"
        # self.slurm_options = " --partition=long --mem=32G"
        self.slurm_options = "--partition=xxl --mem=100G --cpus-per-task=16"

    def set_train_plot_rf_feat_default_options(self):
        self.job_name = "--job-name=RF_importance"
        self.slurm_options = "--partition=short --mem=16G"

    def set_dl1_dl2_default_options(self):
        self.job_name = "--job-name=dl1_2"
        self.slurm_options = "--partition=short --mem=32G"

    def set_dl2_irfs_default_options(self):
        self.job_name = "--job-name=dl2_IRFs"
        self.slurm_partition = "--partition=short"

    def set_dl2_sens_default_options(self):
        self.job_name = "--job-name=dl2_sens"
        self.slurm_options = "--partition=short --mem=32G"

    def set_dl2_sens_plot_default_options(self):
        self.job_name = "--job-name=dl2_sens_plot"
        self.slurm_partition = "--partition=short"



from lstmcpipe.plots.pointings import plot_pointings
import numpy as np
import astropy.units as u
from matplotlib import projections

def test_plot_pointings():
    pointings = np.random.rand(10, 2) * u.rad
    for proj in projections.get_projection_names():
        plot_pointings(pointings, projection=proj, add_grid3d=True, color='black', marker='*', alpha=0.4)



import matplotlib.pyplot as plt
from astropy.table import QTable
import numpy as np
import astropy.units as u
from pyirf.binning import create_bins_per_decade, add_overflow_bins, bin_center
from astropy.io import fits
from astropy.visualization import quantity_support
from ctaplot import ana
import sklearn
from sklearn.multiclass import LabelBinarizer
from matplotlib import ticker


DEFAULT_ENERGY_UNIT = u.TeV
DEFAULT_ENERGY_BINS = create_bins_per_decade(e_min=10 * u.GeV, e_max=100 * u.TeV, bins_per_decade=5)
DEFAULT_THETA2_BINS = np.linspace(0*u.deg**2, 0.4*u.deg**2, num=81)

def format_yaxis_degrees(ax):
    # issue with degrees label formatting see https://github.com/astropy/astropy/issues/13211
    fmt = '{x:.2f}°' if abs(ax.get_ylim()[1] - ax.get_ylim()[0]) < 1 else '{x:.1f}°'
    ax.yaxis.set_major_formatter(ticker.StrMethodFormatter(fmt))
    return ax


class Metric:
    name = 'metric'

    def __init__(self, x, y):
        self.x = x
        self.y = y

    @classmethod
    def from_file(cls, filename):
        table = QTable.read(filename, hdu=cls.name)
        return cls(table['x'], table['y'], table[cls.name])

    def write(self, filename, overwrite=False, append=False):
        """
        Write the metric to a FITS file.

        Parameters
        ----------
        filename: Path | str
            Path to the file to write with the extension .fits.
        overwrite: bool
            If True, overwrite the file if it exists.
        append: bool
            If True, append to the file if it exists, or create it if not.
        """
        hdu = fits.BinTableHDU(self.table, name=self.name)

        if append:
            with fits.open(filename, mode='append') as hdulist:
                hdulist.append(hdu)
        else:
            hdus = [
                fits.PrimaryHDU(),
                hdu
            ]
            fits.HDUList(hdus).writeto(filename, overwrite=overwrite)

    @classmethod
    def from_events(cls, true, reco, x):
        raise NotImplementedError("Should be implemented in child class")

    def plot(self, ax=None, **kwargs):
        ax = plt.gca() if ax is None else ax
        ax.plot(self.x, self.y, **kwargs)
        return ax

    @property
    def table(self):
        return QTable({
            'x': self.x,
            'y': self.y,
        })

class MetricPerBin(Metric):
    name = 'metric'

    def __init__(self, x_low, x_high, y):
        self.x_low = x_low
        self.x_high = x_high
        self.y = y

    @classmethod
    def from_events(cls, true, reco, x, x_bins=10):
        raise NotImplementedError("Should be implemented in child class")

    def table(self):
        return QTable({
            'x_low': self.x_low,
            'x_high': self.x_high,
            'y': self.y,
        })

    @classmethod
    def from_file(cls, filename):
        table = QTable.read(filename, hdu=cls.name)
        return cls(table['x_low'], table['x_high'], table[cls.name])


class MetricPerEnergyBin(Metric):
    name = 'metric_per_energy'

    def __init__(self, energy_bins, metric):
        self._energy_bins = energy_bins.to(DEFAULT_ENERGY_UNIT)
        self._metric = metric

    @property
    def energy_bins(self):
        return self._energy_bins

    @property
    def energy_low(self):
        return self.energy_bins[:-1]

    @property
    def energy_high(self):
        return self.energy_bins[1:]

    @property
    def table(self):
        return QTable({
            'energy_low': self.energy_low,
            'energy_high': self.energy_high,
            self.name: self._metric,
        })

    @property
    def energy_center(self):
        return bin_center(self.energy_bins)

    @classmethod
    def from_file(cls, filename):
        table = QTable.read(filename, hdu=cls.name)
        return cls(table['energy_low'], table['energy_high'], table[cls.name])

    @classmethod
    def from_events(cls, true, reco, energy, energy_bins=DEFAULT_ENERGY_BINS):
        raise NotImplementedError("Should be implemented in child class")

    def plot(self, ax=None, **kwargs):
        ax = plt.gca() if ax is None else ax

        kwargs.setdefault('label', self.name)
        with quantity_support():
            ax.errorbar(self.energy_center,
                        self._metric,
                        xerr=[self.energy_center - self.energy_low, self.energy_high - self.energy_center],
                        **kwargs,
                        )
        ax.set_xscale('log')
        ax.set_xlabel(f'Energy / {self.energy_center.unit}')
        ax.set_title(self.name)
        return ax


class ResolutionPerEnergyBin(MetricPerEnergyBin):
    name = 'resolution'

    def __init__(self, energy_bins, resolution):
        super().__init__(energy_bins, resolution)

    @classmethod
    def from_events(cls, true, reco, energy, energy_bins=DEFAULT_ENERGY_BINS):
        _, res = ana.resolution_per_energy(true, reco, energy, bins=energy_bins)
        return cls(energy_bins, res[:,0])


class AngularResolutionPerEnergyBin(MetricPerEnergyBin):
    name = 'angular_resolution'

    def __init__(self, energy_bins, angular_resolution):
        super().__init__(energy_bins, angular_resolution)

    @classmethod
    def from_events(cls, true_alt, reco_alt, true_az, reco_az, true_energy, energy_bins=DEFAULT_ENERGY_BINS):
        _, res = ana.angular_resolution_per_energy(true_alt, reco_alt, true_az, reco_az, true_energy, bins=energy_bins)
        return cls(energy_bins, res[:,0])

    def plot(self, ax=None, **kwargs):
        ax = super().plot(ax=ax, **kwargs)
        format_yaxis_degrees(ax)
        return ax


class EnergyResolutionPerEnergyBin(MetricPerEnergyBin):
    name = 'energy_resolution'

    def __init__(self, energy_bins, energy_resolution):
        super().__init__(energy_bins, energy_resolution)


    @classmethod
    def from_events(cls, true_energy, reco_energy, energy_bins=DEFAULT_ENERGY_BINS):
        _, res = ana.energy_resolution_per_energy(true_energy, reco_energy, bins=energy_bins)
        return cls(energy_bins, res[:,0])


class ImpactResolutionPerEnergyBin(MetricPerEnergyBin):
        name = 'impact_resolution'

        def __init__(self, energy_bins, impact_resolution):
            super().__init__(energy_bins, impact_resolution)

        @classmethod
        def from_events(cls, true_alt, reco_alt, true_az, reco_az, true_energy, energy_bins=DEFAULT_ENERGY_BINS):
            _, res = ana.impact_resolution_per_energy(true_x, reco_x, true_y, reco_y, true_energy, bins=energy_bins)
            return cls(energy_bins, res[:,0])


class ROCCurve(Metric):
    name = 'roc'

    def __init__(self, fpr, tpr):
        """

        Parameters
        ----------
        fpr: array
            false positive rate
        tpr: array
            true positive rate
        """
        self.tpr = tpr
        self.fpr = fpr

    @classmethod
    def from_events(cls, true_type, reco_proba, pos_label=None, sample_weight=None, drop_intermediate=True):
        auc_score = sklearn.metrics.roc_auc_score(true_type, reco_proba)
        if auc_score < 0.5:
            auc_score = 1 - auc_score

        fpr, tpr, thresholds = sklearn.metrics.roc_curve(true_type,
                                                         reco_proba,
                                                         pos_label=pos_label,
                                                         sample_weight=sample_weight,
                                                         drop_intermediate=drop_intermediate,
                                                         )
        return cls(fpr, tpr)

    @property
    def table(self):
        return QTable({
            'fpr': self.fpr,
            'tpr': self.tpr,
        })


    def plot(self, ax=None, **kwargs):
        ax = plt.gca() if ax is None else ax
        roc_auc = sklearn.metrics.auc(self.fpr, self.tpr)
        rcd = sklearn.metrics.RocCurveDisplay(fpr=self.fpr, tpr=self.tpr, roc_auc=roc_auc)
        rcd.plot(ax=ax, **kwargs)
        return ax

    @classmethod
    def from_file(cls, filename):
        table = QTable.read(filename, hdu=cls.name)
        return cls(table['fpr'], table['tpr'])


class ROCCurveGamma(ROCCurve):
    name = 'roc_gamma'
    def __init__(self, fpr, tpr):
        super().__init__(fpr, tpr)

    def plot(self, ax=None, **kwargs):
        ax = super().plot(ax=ax, **kwargs)
        ax.set_xlabel('')
        ax.set_title("gamma ROC curve")
        ax.set_xlabel("gamma false positive rate")
        ax.set_ylabel("gamma true positive rate")
        return ax

    @classmethod
    def from_events(cls, true_type, gammaness,
                    gamma_label=0,
                    sample_weight=None,
                    drop_intermediate=True):

        from sklearn.preprocessing import label_binarize
        from sklearn.multiclass import OneVsRestClassifier

        if gamma_label is not None and gamma_label in set(true_type):
            label_binarizer = LabelBinarizer()
            binarized_classes = label_binarizer.fit_transform(true_type)

            ii = np.argwhere(label_binarizer.classes_ == gamma_label)[0][0]

            return super().from_events(binarized_classes[:, ii],
                                        gammaness,
                                        pos_label=1,
                                        sample_weight=sample_weight,
                                        drop_intermediate=drop_intermediate)


        else:
            raise ValueError(f"true_type must contain gamma_label {gamma_label}")

    def plot(self, ax=None, **kwargs):
        return super().plot(ax=ax, **kwargs)


class EnergyBias(MetricPerEnergyBin):
    name = 'energy_bias'

    def __init__(self, energy_bins, energy_bias):
        super().__init__(energy_bins, energy_bias)

    @classmethod
    def from_events(cls, true_energy, reco_energy, energy_bins=DEFAULT_ENERGY_BINS):
        _, res = ana.energy_bias_per_energy(true_energy, reco_energy, bins=energy_bins)
        return cls(energy_bins, res[:,0])


class Theta2(Metric):
    name = 'theta2'

    def __init__(self, theta2_bins, theta2_values):
        self.theta2_bins = theta2_bins
        self.theta2_values = theta2_values

    @property
    def theta2_low(self):
        return self.theta2_bins[:-1]

    @property
    def theta2_high(self):
        return self.theta2_bins[1:]

    @classmethod
    def from_events(cls, true_alt, reco_alt, true_az, reco_az, bias_correction=False, bins=DEFAULT_THETA2_BINS):
        from ctaplot.ana.ana import theta2
        ang_sep = theta2(true_alt, reco_alt, true_az, reco_az, bias_correction=bias_correction)
        theta2_values, theta2_bins = np.histogram(ang_sep, bins=bins)
        return cls(theta2_bins=theta2_bins, theta2_values=theta2_values)

    @property
    def table(self):
        return QTable({
            'theta2_low': self.theta2_low,
            'theta2_high': self.theta2_high,
            'theta2_values': self.theta2_values,
        })

    def plot(self, ax=None, **kwargs):
        ax = plt.gca() if ax is None else ax
        with quantity_support():
            ax.stairs(self.theta2_values, edges=self.theta2_bins.to(u.deg**2), **kwargs)
        ax.set_xlabel(r'angular separation ($deg^2$)')
        ax.set_ylabel('count')
        return ax

    @classmethod
    def from_file(cls, filename):
        table = QTable.read(filename, hdu=cls.name)
        theta2_bins = np.append(table['theta2_low'], table['theta2_high'][-1])
        return cls(theta2_bins, table['theta2_values'])


def angular_separation_bias(true_alt, reco_alt, true_az, reco_az):
    from astropy.coordinates import angular_separation
    ang_sep = angular_separation(true_az.mean(), reco_az.mean(), true_alt.mean(), reco_alt.mean())
    return ang_sep

def angular_separation_bias_per_energy(true_alt, reco_alt, true_az, reco_az, true_energy, energy_bins=DEFAULT_ENERGY_BINS):
    biases = []
    for i in range(len(energy_bins) - 1):
        low = energy_bins[i]
        high = energy_bins[i+1]
        mask = (true_energy >= low) & (true_energy < high)
        biases.append(angular_separation_bias(true_az[mask], true_alt[mask], reco_az[mask], reco_alt[mask]))
    return energy_bins, u.Quantity(biases)


class AngularSeparationBiasPerEnergy(MetricPerEnergyBin):
    name = 'angular_separation_bias'

    def __init__(self, energy_low, energy_high, angular_separation_bias):
        super().__init__(energy_low, energy_high, angular_separation_bias.to(u.deg))

    @classmethod
    @u.quantity_input(true_alt=u.rad, reco_alt=u.rad, true_az=u.rad, reco_az=u.rad, true_energy=u.TeV, energy_bins=u.TeV)
    def from_events(cls, true_alt, reco_alt, true_az, reco_az, true_energy, energy_bins=DEFAULT_ENERGY_BINS):
        _, bias = angular_separation_bias_per_energy(true_alt, reco_alt, true_az, reco_az, true_energy, energy_bins=energy_bins)
        return cls(energy_bins[:-1], energy_bins[1:], bias)


    def plot(self, ax=None, **kwargs):
        ax = super().plot(ax=ax, **kwargs)
        ax.set_xlabel(f'Energy / {DEFAULT_ENERGY_UNIT}')
        ax.set_ylabel('angular separation bias')
        ax.set_xscale('log')
        format_yaxis_degrees(ax)
        return ax


class EffectiveArea(MetricPerEnergyBin):
    name = 'effective_area'

    def __init__(self, energy_bins, effective_area):
        super().__init__(energy_bins, effective_area)

    @classmethod
    def from_events(cls, gammas, simulation_info, energy_bins=DEFAULT_ENERGY_BINS):
        from pyirf.irf.effective_area import effective_area_per_energy
        effarea = effective_area_per_energy(gammas, simulation_info, energy_bins)
        return cls(energy_bins, effarea)

    def plot(self, ax=None, **kwargs):
        ax = super().plot(ax=ax, **kwargs)
        ax.set_ylabel(f'Effective area / {self._metric.unit}')
        ax.set_yscale('log')
        ax.grid(True, which='both')
        ax.set_axisbelow(True)
        return ax


class SkyMap(Metric):
    name = 'sky_map'

    def __init__(self, sky_map):
        self.sky_map = sky_map


import mpl_toolkits
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u


def plot_pointings(pointings, ax=None, projection='polar', add_grid3d=False, **kwargs):
    """
    Produce a scatter plot of the pointings

    Parameters
    ----------
    pointings: 2D array of `astropy.quantities` or numpy array in rad
    ax : `matplotlib.pyplot.Axis`
    projection: str or None
        '3d' | 'aitoff' | 'hammer' | 'lambert' | 'mollweide' | 'polar' | 'rectilinear'
    add_grid3d: bool
        add a 3D grid in case of projection='3d'
    kwargs: dict
        kwargs for `matplotlib.pyplot.scatter`

    Returns
    -------
    ax: `matplotlib.pyplot.axis`
    """
    
    if ax and projection:
        if not isinstance(ax, matplotlib.projections.get_projection_class(projection)):
            raise ValueError(f"ax of type {type(ax)} and projection {projection} are exclusive")
        
    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(111, projection=projection)
    
    elif isinstance(ax, mpl_toolkits.mplot3d.axes3d.Axes3D):
        projection = '3d'
    elif isinstance(ax, matplotlib.projections.polar.PolarAxes):
        projection = 'polar'
        
    if projection == '3d':
        r = 1.
        if add_grid3d:
            az, alt = np.mgrid[0:2.01*np.pi:(1/10)* np.pi/2., 0:np.pi/2.:(1/10)* np.pi/4.]
            X = r * np.cos(az) * np.cos(alt)
            Y = r * np.sin(az) * np.cos(alt)
            Z = r * np.sin(alt)
            ax.plot_surface(X, Y, Z, cmap=plt.cm.YlGnBu_r, alpha=.1)
        r *= 1.01
        az = pointings[:,0]
        alt = pointings[:,1]
        X = r * np.cos(az) * np.cos(alt)
        Y = r * np.sin(az) * np.cos(alt)
        Z = r * np.sin(alt)

        ax.scatter(X, Y, Z, **kwargs)

        box_ratio=1.03
        ax.set_xlim3d([-r*box_ratio, r*box_ratio])
        ax.set_ylim3d([-r*box_ratio, r*box_ratio])
        ax.set_zlim3d([-r*box_ratio, r*box_ratio])
        
    elif projection == 'polar':
        ax.scatter(pointings[:, 0], np.pi/2.*u.rad - pointings[:, 1], **kwargs)
        ax.set_xlabel('Azimuth')
        rticks_deg = [10, 30, 50, 70, 90]
        ax.set_rticks(np.deg2rad(rticks_deg), labels=[f'{r}°' for r in rticks_deg])
        ax.set_rmax(np.pi/2.)
        ax.set_rlabel_position(20)
        
        
    else:
        ax.scatter(pointings[:, 0], pointings[:, 1], **kwargs)
        ax.set_xlabel('Azimuth')
        ax.set_ylabel('Altitude')
        
        
    ax.legend()
    ax.grid(True)
    ax.set_axisbelow(True)

    return ax




from pathlib import Path
import yaml
import git
from tabulate import tabulate


def load_config(filename):
    with open(filename, 'r') as f:
        return yaml.safe_load(f)


def add_prod_table(
    production_dir,
    prod_file='productions.rst',
    lstmcpipe_repo_prod_config_url='https://github.com/cta-observatory/lstmcpipe/tree/master/production_configs/',
):
    """

    Parameters
    ----------
    production_dir: pathlib.Path or str
        path to the production configs directory
    prod_file: str
        RST prod file to update
    lstmcpipe_repo_prod_config_url: str
    """

    prod_txt = '\n'

    commit_times = []
    prod_list = []
    for prod_dir in [d for d in Path(production_dir).iterdir() if d.is_dir()]:
        commit = list(git.Repo(root_dir).iter_commits(paths=prod_dir, max_count=1))[0]

        yml_list = [f for f in prod_dir.iterdir() if f.name.endswith('.yml') or f.name.endswith('.yaml')]
        if yml_list:
            try:
                conf = load_config(yml_list[0])
                prod_id = conf['prod_id']
            except:  # noqa
                print(f"Could not load prod id for {prod_dir.name}")
                prod_id = ''
        else:
            print(f"No yml file in {prod_dir}")

        markdowns = list(prod_dir.glob('*.md'))
        readme = open(markdowns[0]).read() if markdowns else ""
        prod_list.append(
            [
                commit.authored_datetime.date(),
                f"`{prod_dir.name} <{lstmcpipe_repo_prod_config_url+prod_dir.name}>`_",
                prod_id,
                readme,
            ]
        )
        commit_times.append(commit.authored_datetime)

    sorted_lists = sorted(zip(commit_times, prod_list))
    prod_list = [prod for _, prod in sorted_lists]

    prod_txt += tabulate(prod_list, ['Request date', 'Directory name', 'Prod ID', 'readme'], tablefmt='rst')

    with open(prod_file, 'a') as f:
        f.write(prod_txt)


if __name__ == '__main__':
    root_dir = Path(__file__).parent.joinpath('..').resolve()
    prod_dir = root_dir.joinpath('production_configs')

    add_prod_table(prod_dir)



import json
import pytest
import tempfile
from pathlib import Path
from ruamel.yaml import YAML
from lstmcpipe.utils import rerun_cmd, dump_lstchain_std_config, SbatchLstMCStage, run_command


def test_save_log_to_file():
    from ..utils import save_log_to_file

    outfile_yml = Path('dummy_log.yml')

    dummy_log = {'dummy_jobid': 'sbatch --parsable --wrap="sleep 10"'}
    save_log_to_file(dummy_log, outfile_yml)

    assert outfile_yml.exists()
    with open(outfile_yml) as f:
        log = YAML().load(f)

    assert "NoKEY" in log.keys()
    assert isinstance(log["NoKEY"], dict)
    assert "dummy_jobid" in log["NoKEY"].keys()
    assert dummy_log["dummy_jobid"] == log["NoKEY"]["dummy_jobid"]


def test_rerun_cmd():

    with tempfile.TemporaryDirectory() as tmp_dir:
        file, filename = tempfile.mkstemp(dir=tmp_dir)
        cmd = f'echo "1" >> {filename}; rm nonexistingfile'
        # first test: the cmd fails 3 times but the outfile stays in place
        subdir_failures = ''
        rerun_cmd(cmd, filename, max_ntry=3, subdir_failures=subdir_failures, shell=True)
        filename = Path(filename)
        filename = Path(tmp_dir).joinpath(subdir_failures, filename.name)
        assert open(filename).read() == "1\n1\n1\n"
        # 2nd test: the cmd fails and the outfile is moved in subdir
        subdir_failures = 'fail'
        rerun_cmd(cmd, filename, max_ntry=3, subdir_failures=subdir_failures, shell=True)
        filename = filename.parent.joinpath(subdir_failures).joinpath(filename.name)
        assert open(filename).read() == "1\n"
        assert filename.exists()


def test_rerun_cmd_lstchain_mc_r0_to_dl1():
    with tempfile.TemporaryDirectory() as tmp_dir:
        cmd = ['lstchain_mc_r0_to_dl1', '-o', tmp_dir]
        outfile = Path(tmp_dir, 'dl1_gamma_test_large.h5')
        # first try should succeed
        ntry = rerun_cmd(cmd, outfile, max_ntry=3)
        assert ntry == 1
        # second try should fail because the outfile already exists
        ntry = rerun_cmd(cmd, outfile, max_ntry=3, subdir_failures='failed_outputs')
        assert ntry == 2
        assert Path(tmp_dir, 'failed_outputs', 'dl1_gamma_test_large.h5').exists()


def test_run_command():
    cmd = run_command("echo 'this is the command to be passed'")
    assert cmd == "this is the command to be passed"


@pytest.mark.xfail(raises=ValueError)
def test_fail_run_command():
    run_command("echoo 'this command will fail'")


def test_sbatch_lst_mc_stage():
    with pytest.raises(TypeError):
        SbatchLstMCStage()
        SbatchLstMCStage("merge")

    sbatch = SbatchLstMCStage(stage="r0_to_dl1", wrap_command="command to be batched")
    assert (
        sbatch.slurm_command == 'sbatch --parsable --job-name=r0_dl1 --partition=long --array=0-0%100  '
        '--error=./slurm-%j.e --output=./slurm-%j.o   --wrap="command to be batched"'
    )

    sbatch.slurm_options = "-A lstrta -p xxl --mem 160G --cpus-per-task=32"
    assert (
        sbatch.slurm_command == 'sbatch --parsable --job-name=r0_dl1 -A lstrta -p xxl --mem 160G --cpus-per-task=32 '
        '--error=./slurm-%j.e --output=./slurm-%j.o   --wrap="command to be batched"'
    )

    sbatch.slurm_options = None
    sbatch.check_slurm_dependencies("123,243,345,456")
    assert sbatch.slurm_dependencies == "--dependency=afterok:123,243,345,456"

    sbatch.compose_wrap_command(
        wrap_command="python args",
        source_env="source .bashrc_file; conda activate   ",
        backend="  export MPLBACKEND=Agg    ",
    )
    assert sbatch.wrap_cmd == '--wrap="export MPLBACKEND=Agg; source .bashrc_file; conda activate; python args"'

    with pytest.raises(ValueError):
        sbatch.submit()  # slurm not installed
        sbatch.stage_default_options("invented_stage")
        sbatch.check_slurm_dependencies(slurm_deps="123,,234")

    stages = sbatch._valid_stages
    for stage in stages:
        sbatch.stage_default_options(stage)
        assert "--job-name=" in sbatch.job_name
        assert "--partition=" in sbatch.slurm_command


def test_dump_lstchain_std_config():
    with tempfile.TemporaryDirectory() as tmpdir:
        outfile = Path(tmpdir).joinpath('cfg.json')
        dump_lstchain_std_config(filename=outfile, allsky=False)
        assert json.load(outfile.open())['GlobalPeakWindowSum']['apply_integration_correction']
        dump_lstchain_std_config(filename=outfile, allsky=True, overwrite=True)
        assert 'alt_tel' in json.load(outfile.open())['energy_regression_features']



#!/usr/bin/env python

# clean directory for DL1 files with wrong parameters table shape
# see https://github.com/cta-observatory/cta-lstchain/issues/978
# simply delete files with wrong table shape


import logging
from pathlib import Path
from ctapipe.io import read_table
from lstchain.io.io import dl1_params_lstcam_key


def clean_dir(root_dir, ncols=56):
    root_dir = Path(root_dir)
    dl1_files = root_dir.rglob('dl1_*.h5')
    for file in dl1_files:
        param_table = read_table(file, dl1_params_lstcam_key)
        if param_table.to_pandas().shape[1] < ncols:
            logging.warning(f"File {file} will be deleted, its param table shape is {param_table.to_pandas().shape}")
            file.unlink()


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Clean dir for faulty DL1 files')
    parser.add_argument('directory', type=Path, help='Input directory')
    parser.add_argument(
        '--ncols', dest='ncols', type=int, default=56, help='Number of expected columns in DL1 parameters table'
    )

    args = parser.parse_args()

    logging.basicConfig(filename='cleandir.log', level=logging.DEBUG)
    logging.info(f'Cleaning directory {args.directory}\n\n')

    clean_dir(args.directory, args.ncols)



# Try to use setuptools_scm to get the current version; this is only used
# in development installations from the git repository.
# see lstmcpipe/version.py for details
from .. import lstmcpipe_root_dir

try:
    from setuptools_scm import get_version

    version = get_version(root=lstmcpipe_root_dir)
except Exception as e:
    raise ImportError(f"setuptools_scm broken or not installed: {e}")



#!/usr/bin/env python


import subprocess as sp
from pathlib import Path
from lstmcpipe.config.paths_config import PathConfigAllSkyFull

prod_id = '20220511_allsky_std'
list_decs = ['dec_2276', 'dec_3476', 'dec_4822', 'dec_931', 'dec_min_413']
source_env = "source /fefs/aswg/software/conda/etc/profile.d/conda.sh; conda activate lstchain-v0.9.6; "

pcf = PathConfigAllSkyFull(prod_id, list_decs)


def find_train_dirs(config):
    dirs = []
    for dec, train in config.train_configs.items():
        for particle in train.training_particles:
            for pointing in train.pointing_dirs(particle):
                dirs.append(train.dl1_dir(particle, pointing))

    return dirs


def find_test_dirs(config):
    dirs = []
    for dec, test in config.test_configs.items():
        for pointing in test.pointing_dirs():
            dirs.append(test.dl1_dir(pointing))

    return dirs


if __name__ == '__main__':
    train_dirs = find_train_dirs(pcf)
    test_dirs = find_test_dirs(pcf)

    for dir_i in train_dirs + test_dirs:
        if not Path(dir_i).exists():
            continue

        node = dir_i.split('/')[-1]

        error_file = f'./clean_dir_{prod_id}_{node}.e'
        output_file = f'./clean_dir_{prod_id}_{node}.o'

        cmd = f'{source_env} python ./cleandir.py {dir_i}'

        slurm_cmd = f'sbatch --parsable --error={error_file} --output={output_file} --wrap="{cmd}"'

        run_command = sp.Popen(slurm_cmd, shell=True, stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf-8')
        stdout, _ = run_command.communicate()

        if run_command.returncode != 0:
            raise ValueError(
                f"Running {run_command} failed with return code {run_command.returncode}" f", output:" f" \n {stdout}"
            )



from lstmcpipe.config.paths_config import PathConfigAllSkyFull

prod_id = '20220511_allsky_std'

pcf = PathConfigAllSkyFull(prod_id,
        ['dec_2276','dec_3476','dec_4822','dec_931','dec_min_413'])

pcf.generate()

pcf.save_yml(f'lstmcpipe_config_{prod_id}.yml')


from lstchain.io.config import get_standard_config
import json
filename = f'lstchain_config_{prod_id}.json'

cfg = get_standard_config()
cfg['LocalPeakWindowSum']['apply_integration_correction'] = True
cfg['GlobalPeakWindowSum']['apply_integration_correction'] = True
cfg['source_config']['EventSource']['allowed_tels'] = [1]
cfg['random_forest_energy_regressor_args']['min_samples_leaf'] = 10
cfg['random_forest_disp_regressor_args']['min_samples_leaf'] = 10
cfg['random_forest_disp_classifier_args']['min_samples_leaf'] = 10
cfg['random_forest_particle_classifier_args']['min_samples_leaf'] = 10
with open(filename, 'w') as file:
    json.dump(cfg, file, indent=4)

print(f"Modified lstchain config dumped in {filename}")



"""
Script to generate a mini production tree with symlinks to a few simtel files and the corresponding path config.
This is used to test lstmcpipe on a small prod
"""

from pathlib import Path
from tqdm import tqdm
from datetime import date
import os
import argparse

from lstmcpipe.config import paths_config
from lstmcpipe.utils import dump_lstchain_std_config


def generate_tree(base_dir, working_dir, nfiles):
    """
    Walk the base dir looking for simtels files
    When a directory contains simtels files, it's tree structure is duplicated into the working dir
    and nfiles are symlinked there
    """
    base_dir = Path(base_dir)
    working_dir = Path(working_dir)

    for root, dirs, files in tqdm(os.walk(base_dir)):
        simtel_files = [os.path.join(root, file) for file in files if file.endswith('.simtel.gz')]
        if simtel_files:
            for file in simtel_files[:nfiles]:
                target = working_dir.joinpath(Path(file).relative_to(base_dir))
                target.parent.mkdir(parents=True, exist_ok=True)
                if not target.exists():
                    target.symlink_to(file)


def generate_test_prod5trans80(
    working_dir, nfiles=5, path_to_config_file='.', overwrite=True
):
    base_dir = '/fefs/aswg/workspace/lstmcpipe/data/test_data/mc/DL0/20200629_prod5_trans_80/'
    working_dir = os.path.join(working_dir, 'DL0/20200629_prod5_trans_80/')

    generate_tree(base_dir, working_dir, nfiles)

    pc = paths_config.PathConfigProd5Trans80(f'test_prod_{date.today()}')
    pc.base_dir = os.path.join(
        working_dir, '{data_level}/20200629_prod5_trans_80/{particle}/{zenith}/south_pointing/{prod_id}'
    )
    pc.generate()
    pc.save_yml(os.path.join(path_to_config_file, f'test_prod5trans80_{date.today()}.yaml'), overwrite=overwrite)


def generate_test_allsky(
    working_dir,
    nfiles=5,
    path_to_config_file='.',
    decs=['dec_4822', 'dec_931'],
    overwrite=True,
):
    allsky_train_base_dir = '/home/georgios.voutsinas/ws/AllSky'
    allsky_test_base_dir = '/home/georgios.voutsinas/ws/AllSky'

    generate_tree(allsky_train_base_dir, os.path.join(working_dir, 'DL0/AllSky'), nfiles)

    pc = paths_config.PathConfigAllSkyFull(f'test_prod_{date.today()}', decs)
    pcdl1ab = paths_config.PathConfigAllSkyFullDL1ab(f'test_prod_{date.today()}', f'test_prod_{date.today()}_dl1ab', decs, run_checker=False)
    
    # config training dir are replaced with local ones
    for dec in decs:
        pc.train_configs[dec].base_dir = os.path.join(
            working_dir, '{data_level}/AllSky/{prod_id}/{dataset_type}/{dec}/{particle}/{pointing}/'
        )
        pc.test_configs[dec].base_dir = os.path.join(
            working_dir, '{data_level}/AllSky/{prod_id}/{dataset_type}/{dec}/{particle}/{pointing}/'
        )
        pc.train_configs[dec].training_dir = os.path.join(
            working_dir, pc.train_configs[dec].training_dir.replace(allsky_train_base_dir, 'DL0/AllSky/')
        )
        pc.test_configs[dec].testing_dir = os.path.join(
            working_dir, pc.test_configs[dec].testing_dir.replace(allsky_test_base_dir, 'DL0/AllSky/')
        )
        pcdl1ab.train_configs[dec].base_dir = os.path.join(
            working_dir, '{data_level}/AllSky/{prod_id}/{dataset_type}/{dec}/{particle}/{pointing}/'
        )
        pcdl1ab.train_configs[dec].source_config.base_dir = os.path.join(
            working_dir, '{data_level}/AllSky/{prod_id}/{dataset_type}/{dec}/{particle}/{pointing}/'
        )
        pcdl1ab.test_configs[dec].base_dir = os.path.join(
            working_dir, '{data_level}/AllSky/{prod_id}/{dataset_type}/{dec}/{particle}/{pointing}/'
        )
        pcdl1ab.test_configs[dec].source_config.base_dir = os.path.join(
            working_dir, '{data_level}/AllSky/{prod_id}/{dataset_type}/{dec}/{particle}/{pointing}/'
        )
        pcdl1ab.train_configs[dec].training_dir = os.path.join(
            working_dir, pc.train_configs[dec].training_dir.replace(allsky_train_base_dir, 'DL0/AllSky/')
        )
        pcdl1ab.train_configs[dec].source_config.training_dir = os.path.join(
            working_dir, pc.train_configs[dec].training_dir.replace(allsky_train_base_dir, 'DL0/AllSky/')
        )
        pcdl1ab.test_configs[dec].testing_dir = os.path.join(
            working_dir, pc.test_configs[dec].testing_dir.replace(allsky_test_base_dir, 'DL0/AllSky/')
        )
        pcdl1ab.test_configs[dec].source_config.testing_dir = os.path.join(
            working_dir, pc.test_configs[dec].testing_dir.replace(allsky_test_base_dir, 'DL0/AllSky/')
        )
        
    pc.generate()
    for stage_name, stage_steps in pc.paths.items():
        for step in stage_steps:
            step['slurm_options'] = '-p short'
    pc.save_yml(os.path.join(path_to_config_file, f'test_AllSky_{date.today()}.yaml'), overwrite=overwrite)
    
    pcdl1ab.generate()
    for stage_name, stage_steps in pcdl1ab.paths.items():
        for step in stage_steps:
            step['slurm_options'] = '-p short'
    pcdl1ab.save_yml(os.path.join(path_to_config_file, f'test_AllSky_{date.today()}_dl1ab.yaml'), overwrite=overwrite)



if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Generate test tree')

    parser.add_argument('prod_type', type=str, help='prod5trans80 or allsky')
    parser.add_argument('--nfiles', type=int, default=2, help='Number of files')
    parser.add_argument('--path_config_file', type=Path, default='.', help='Path to save the corresponding config file')
    parser.add_argument(
        '--working_dir',
        type=Path,
        default='/fefs/aswg/workspace/lstmcpipe/data/mc/',
        help='Your working dir where the DL0 tree will be generated, '
        '(such as /fefs/aswg/workspace/firstname.surname/data/mc/)',
    )

    args = parser.parse_args()

    if args.working_dir is None:
        user = os.environ['USER']
        if user == 'lstanalyzer':
            working_dir = '/fefs/aswg/workspace/lstmcpipe/data/mc'
        else:
            working_dir = f'/fefs/aswg/workspace/{user}/data/mc'
        if not Path(working_dir).exists():
            raise FileNotFoundError(f"working dir {working_dir} does not exist, provide one")
    else:
        working_dir = args.working_dir

    if args.prod_type == 'prod5trans80':
        generate_test_prod5trans80(working_dir, args.nfiles, args.path_config_file)
    elif args.prod_type == 'allsky':
        generate_test_allsky(working_dir, args.nfiles, args.path_config_file)
    else:
        raise NotImplementedError("Unknown prod type")
        
    dump_lstchain_std_config(filename=Path(args.path_config_file, 'lstchain_config.json'), overwrite=True)



import pytest

from lstmcpipe.io.data_management import check_data_path, get_input_filelist


@pytest.fixture()
def create_tmp_dir(tmp_path):
    test_dir = tmp_path / 'data'
    test_dir.mkdir()
    return test_dir


@pytest.fixture()
def create_tmp_subdir(tmp_path):
    test_dir = tmp_path / 'data' / 'subdir'
    test_dir.mkdir()
    return test_dir


@pytest.mark.xfail(raises=ValueError)
def test_check_data_path_raise_value_error(create_tmp_dir):
    check_data_path(create_tmp_dir)


def test_get_input_filelist(create_tmp_dir, create_tmp_subdir):
    for i in range(2):
        with open(create_tmp_dir.joinpath(f'LST-1.1.Run00101.{i:05d}.simtel.fz'), 'w'):
            pass
        with open(create_tmp_dir.joinpath(f'LST-1.1.Run00102.{i:05d}.h5'), 'w'):
            pass
    with open(create_tmp_dir.joinpath('some_file.log'), 'w'):
        pass
    for i in range(2):
        with open(create_tmp_subdir.joinpath(f'LST-1.1.Run00103.{i:05d}.simtel.fz'), 'w'):
            pass

    assert len(get_input_filelist(create_tmp_dir)) == 6  # 5 files plus subdir
    assert len(get_input_filelist(create_tmp_dir, glob_pattern="*.h5")) == 2
    assert len(get_input_filelist(create_tmp_dir, glob_pattern="**/*.simtel.fz")) == 4



# #!/usr/bin/env python
import argparse
from pathlib import Path
import tempfile
import subprocess

from lstmcpipe.config.pipeline_config import load_config
from lstchain.io.config import read_configuration_file
from ctapipe.utils import get_dataset_path


def mc_r0_path():
    return get_dataset_path("gamma_test_large.simtel.gz")


def run_lstchain_mc_r0_dl1(simtel_file, outdir, config_file):
    cmd = ['lstchain_mc_r0_to_dl1', '-f', simtel_file, '-o', outdir, '-c', config_file]
    return subprocess.run(cmd)


def test_r0_dl1(config_file):
    tmpdir = tempfile.mkdtemp()
    run_lstchain_mc_r0_dl1(mc_r0_path(), tmpdir, config_file)


def test_dl1ab(config_file):
    tmpdir = tempfile.mkdtemp()
    dl1file = Path(tmpdir, 'dl1_gamma_test_large.h5')
    dl1fileout = Path(tmpdir, 'dl1ab.h5')
    run_lstchain_mc_r0_dl1(mc_r0_path(), tmpdir, config_file)
    cmd = ['lstchain_dl1ab', '-f', dl1file, '-o', dl1fileout, '-c', config_file]
    subprocess.run(cmd)


def validate_lstchain(config_file, stage):
    if stage == 'r0_to_dl1':
        test_r0_dl1(config_file)
    elif stage == 'dl1ab':
        test_dl1ab(config_file)
    else:
        # else we simply validate that reading the config is ok
        read_configuration_file(config_file)


def main():
    parser = argparse.ArgumentParser(description="Validate your lstmcpipe config")

    # Required arguments
    parser.add_argument(
        dest='lstmcpipe_config_filename',
        type=Path,
        help='Path to lstmcpipe config file',
    )

    # optional
    parser.add_argument(
        '--lstchain_config',
        '-lc',
        action='store',
        type=Path,
        dest='lstchain_config_filename',
        help='Path to the lstchain config file.',
        default=None,
    )

    args = parser.parse_args()

    config = load_config(args.lstmcpipe_config_filename)
    if args.lstchain_config_filename is not None:
        validate_lstchain(args.lstchain_config_filename, stage=config['stages_to_run'][0])

    print("\n\nValid configs :)")


if __name__ == '__main__':
    main()




#!/usr/bin/env python

import shutil
import logging
from pathlib import Path
from ..utils import save_log_to_file, SbatchLstMCStage
from ..io.data_management import check_and_make_dir_without_verification


log = logging.getLogger(__name__)


def batch_train_pipe(dict_paths, jobids_from_merge, config_file, batch_config, logs):
    """
    Function to batch the lstchain train_pipe once the proton and gamma-diffuse merge_and_copy_dl1 batched jobs have
    finished.

    Parameters
    ----------
    dict_paths : dict
        Core dictionary with {stage: PATHS} information
    config_file : str
        Path to a configuration file. If none is given, a standard configuration is applied
    jobids_from_merge : str
        string containing the jobids (***ONLY from proton and gamma-diffuse***) from the jobs batched in the
         merge_and_copy_dl1 stage, to be passed to the train_pipe function (as a slurm dependency)
    batch_config : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to
        the `train_pipe` function.
    logs: dict
        Dictionary with logs files

    Returns
    -------
    jobid_4_dl1_to_dl2 : str
        string containing the jobid to be passed to the next stage of the workflow (as a slurm dependency).
        For the next stage, however, it will be needed TRAIN + MERGED jobs
    """
    log_train = {}
    debug_train = {}
    jobid_for_dl1_to_dl2 = []

    log.info("==== START {} ====".format("batch mc_train_workflow"))

    for paths in dict_paths:

        gamma_dl1_train_file = paths["input"]["gamma"]
        proton_dl1_train_file = paths["input"]["proton"]
        models_dir = paths["output"]

        job_logs, jobid = train_pipe(
            gamma_dl1_train_file,
            proton_dl1_train_file,
            models_dir,
            config_file=config_file,
            batch_configuration=batch_config,
            wait_jobs_dl1=jobids_from_merge,
            slurm_options=paths.get("slurm_options", None),
        )

        log_train.update(job_logs)
        jobid_for_dl1_to_dl2.append(jobid)

        debug_train[jobid] = (
            f"The single jobid from train_pipe that depends of {jobids_from_merge} - merge" f"_and_copy jobids"
        )

    jobid_for_dl1_to_dl2 = ",".join(jobid_for_dl1_to_dl2)

    save_log_to_file(log_train, logs["log_file"], workflow_step="train_pipe")
    save_log_to_file(debug_train, logs["debug_file"], workflow_step="train_pipe")

    log.info("==== END {} ====".format("batch mc_train_workflow"))

    return jobid_for_dl1_to_dl2


def batch_plot_rf_features(
    dict_paths,
    config_file,
    batch_configuration,
    train_jobid,
    logs,
):
    """
    Batches the plot_model_importance.py script that creates a .png with the RF feature's importance models
    after the RF are trained.
    The plot is saved in the same dir in where the modes are stored.

    Parameters
    ----------
     dict_paths : dict
        Core dictionary with {stage: PATHS} information
    config_file: str
        Path to lstchain config file
    batch_configuration : dict
        Dictionary containing the (full) source_environment and the slurm_account strings.
    train_jobid: str
        Single jobid from training stage.
    logs: dict
        Dictionary with logs files

    Returns
    -------
    jobid : str
        jobid of batched stage
    """
    log_rf_feat = {}
    log_debug = {}
    all_jobs_plot_rf_feat = []

    log.info("==== START {} ====".format("batch plot RF features importance"))

    for path in dict_paths:
        models_dir = path["output"]

        cmd = f"lstmcpipe_plot_models_importance {models_dir} -cf {config_file}"

        sbatch_rf_feat = SbatchLstMCStage(
            "RF_importance",
            wrap_command=cmd,
            slurm_error=Path(models_dir).joinpath("job_plot_rf_feat_importance.e").resolve().as_posix(),
            slurm_output=Path(models_dir).joinpath(models_dir, "job_plot_rf_feat_importance.o").resolve().as_posix(),
            slurm_deps=train_jobid,
            slurm_account=batch_configuration["slurm_account"],
            source_environment=batch_configuration["source_environment"],
            backend="export MPLBACKEND=Agg;",
        )

        jobid = sbatch_rf_feat.submit()

        log_rf_feat.update({jobid: sbatch_rf_feat.slurm_command})
        log_debug[jobid] = "Single job_id to plot RF feature s importance"
        all_jobs_plot_rf_feat.append(jobid)

    all_jobs_plot_rf_feat = ','.join(all_jobs_plot_rf_feat)

    save_log_to_file(log_rf_feat, logs["log_file"], workflow_step="plot_RF_features_importance")
    save_log_to_file(log_debug, logs["debug_file"], workflow_step="plot_RF_features_importance")

    log.info(" Random Forest importance's plot will be saved at: {}".format(models_dir))
    log.info("==== END {} ====".format("batch plot RF features importance"))

    return all_jobs_plot_rf_feat


def train_pipe(
    gamma_dl1_train_file,
    proton_dl1_train_file,
    models_dir,
    config_file=None,
    batch_configuration='',
    wait_jobs_dl1=None,
    slurm_options=None,
):
    """
    Train RF from MC DL1 data (onsite LaPalma cluster)

    Parameters
    ----------
    gamma_dl1_train_file: str
        FILE path to the gamma
    proton_dl1_train_file: str
        FILE path to the proton
    models_dir:
        DIR trained models output path
    config_file: str
        Path to a configuration file. If none is given, a standard configuration is applied
    batch_configuration : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to the
        sbatch commands
        ! NOTE : train_pipe AND dl1_to_dl2 MUST BE RUN WITH THE SAME ENVIRONMENT
    wait_jobs_dl1 : str
        A string (of chained job_ids separated by ',' and without spaces between each element), containing
        all the job_ids of the merging stage
    slurm_options: str
        Extra slurm options to be passed to the sbatch command

    Returns
    -------
    log_train : dict
        dictionary containing the jobid of the batched job as key and the run command (the lstchain_mc_train
        command with all its corresponding arguments) as value.

    jobid_train : str
        jobid of the batched job to be send (for dependencies purposes) to the next stage of the
        workflow (onsite_mc_dl1_to_dl2)
    """
    log_train = {}

    log.info("Models will be placed in {}".format(models_dir))
    check_and_make_dir_without_verification(models_dir)

    cmd = f"lstchain_mc_trainpipe --fg {gamma_dl1_train_file} --fp {proton_dl1_train_file} -o {models_dir}"
    if config_file is not None:
        cmd = cmd + " -c {}".format(config_file)

    sbatch_train_pipe = SbatchLstMCStage(
        "train_pipe",
        wrap_command=cmd,
        slurm_error=Path(models_dir).joinpath("train_job.e").resolve().as_posix(),
        slurm_output=Path(models_dir).joinpath("train_job.o").resolve().as_posix(),
        slurm_deps=wait_jobs_dl1,
        slurm_options=slurm_options,
        slurm_account=batch_configuration["slurm_account"],
        source_environment=batch_configuration["source_environment"],
    )

    jobid_train = sbatch_train_pipe.submit()
    log_train.update({jobid_train: sbatch_train_pipe.slurm_command})

    log.info(f"Submitted batch job {jobid_train}")

    # copy config into working dir
    if config_file is not None:
        shutil.copyfile(config_file, Path(models_dir).joinpath(Path(config_file).name))

    return log_train, jobid_train
